##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Rob Miller <rob@kalistra.com> (RaFromBRC)
#   - and contributors
#
##########################################################
"""
WickedDoc
~~~~~~~~~~

A simple subclass of the ATDocument type that supports wicked
linking in the primary text field.

"""
from AccessControl import ClassSecurityInfo
from Products.Archetypes import public as atapi
from Products.CMFCore import permissions as CMFCorePermissions
from Products.CMFCore.utils import getToolByName
from wicked.at.field import WikiField
from wicked.atcontent import zope2
from wicked.interfaces import IAmWicked
from zope.annotation.interfaces import IAttributeAnnotatable 
from zope.interface import implements

try:
    from Products.ATContentTypes.atct import ATDocument
except ImportError: # ATCT 0.2
    from Products.ATContentTypes.types.ATDocument import ATDocument
try:
    from Products.ATContentTypes.config import zconf
    ATDOCUMENT_CONTENT_TYPE = zconf.ATDocument.default_content_type
except ImportError: # ATCT 0.2
    from Products.ATContentTypes.config import ATDOCUMENT_CONTENT_TYPE

     
class WickedDoc(ATDocument):
    """ A page in the portal; supports wiki linking. """
    implements(IAmWicked, IAttributeAnnotatable)
    archetype_name='Wicked Doc'
    portal_type= meta_type ='WickedDoc'
    global_allow=True
    schema = ATDocument.schema.copy() + atapi.Schema((
        WikiField('text',
                  required=True,
                  searchable=True,
                  primary=True,
                  filters=('Wicked Filter',),
                  validators = ('isTidyHtmlWithCleanup',),
                  default_content_type = ATDOCUMENT_CONTENT_TYPE,
                  default_output_type = 'text/html',
                  allowable_content_types = ('text/structured',
                                             'text/x-rst',
                                             'text/html',
                                             'text/plain',
                                             'text/plain-pre',
                                             'text/python-source',),
                  widget = atapi.RichWidget(description = "The body text "\
                                            "of the document.",
                                            description_msgid = "help_body_text",
                                            label = "Body text",
                                            label_msgid = "label_body_text",
                                            rows = 25,
                                            i18n_domain = "plone")),
        ))

atapi.registerType(WickedDoc, zope2.PROJECTNAME)



from Products.AdvancedQuery import Eq, Generic
from Products.CMFCore.utils import getToolByName
from wicked.interfaces import IWickedQuery, IAmWicked
from wicked.utils import memoizedproperty, memoize, match, packBrain, cleanUID
from zope.interface import implements
from zope.component import adapts


_marker = object()
class AdvQueryMatchingSeeker(object):
    """
    An advanced query specific 
    CMFish catalog query handler
    """
    implements(IWickedQuery)
    adapts(IAmWicked)

    chunk = _marker
    normalled = _marker
    scope = _marker
    
    def __init__(self, context):
        self.context = context
        self.catalog = getToolByName(context, 'portal_catalog')
        self.path = '/'.join(context.aq_inner.aq_parent.getPhysicalPath())    
        self.evalQ = self.catalog.evalAdvancedQuery

    def configure(self, chunk, normalled, scope):
        self.chunk = chunk
        self.normalled = normalled
        self.scope = scope

    def _query(self, query, sort=('created',)):
        if sort:
            return self.evalQ(query, sort)
        else:
            return self.evalQ(query)

    def queryUIDs(self, uids):
        return self._query(Generic('UID', uids), sort=None)

    @property
    def scopedQuery(self):
        chunk, title = self.chunk, self.title
        query = (Eq('getId', chunk) | Eq('Title', title))
        if not self.scope is _marker:
            # XXX let's move this out of attr storage
            # on the content to at least an annotation
            scope = getattr(self.context, self.scope, self.scope)
            if callable(scope):
                scope = scope()
            if scope:
                query = Generic('path', scope) & query
        return query

    @property
    def basicQuery(self):
        chunk, normalled = self.chunk, self.normalled
        getId = chunk
        self.title = title = '"%s"' % chunk
        query = Generic('path', {'query': self.path, 'depth': -1}) \
                & (Eq('getId', chunk) | Eq('Title', title) | Eq('getId', normalled))
        return query

    @property
    @match
    def scopedSearch(self):
        return self._query(self.scopedQuery)

    @property
    @match
    def search(self):
        return self._query(self.basicQuery)

    def _aggquery(self, name, query):
        curr = getattr(self, name, _marker)
        if curr is _marker:
            curr = query
        else:
            curr |= query
        setattr(self, name, curr)
        return curr

    @property
    def bquery(self):
        return self._aggquery('_bquery', self.basicQuery)

    @property
    def squery(self):
        return self._aggquery('_squery', self.scopedQuery)

    # memo prevents dups
    @memoize 
    def aggregate(self, link, normalled, scope):
        """
        builds aggregated queries for scoped and basic
        """
        self.configure(link, normalled, scope)
        self.bquery 
        self.squery 

    @memoizedproperty
    def agg_brains(self):
        """
        aggregregate search returns
        """
        return list(self._query(self._bquery))

    @memoizedproperty
    def agg_scoped_brains(self):
        """
        aggregregate search returns
        """
        return list(self._query(self._squery))

    __call__ = _query



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from persistent import Persistent
from persistent.mapping import PersistentMapping
from BTrees.OOBTree import OOBTree
from wicked.interfaces import ICacheManager, IAmWicked, IWickedFilter
from wicked.interfaces import IUID
from zope.annotation.interfaces import IAnnotations, IAnnotatable
from zope.interface import implements
from zope.component import adapts, getUtility
from utils import memoizedproperty, memoize, clearbefore

_marker = object()

# @@ need sys-module alias

class CacheStore(Persistent):
    """
    basic persistent cache object
    
    see cache.txt
    """
    def __init__(self, id_):
        self.id = id_
        super(CacheStore, self).__init__(self)
        self.field={}
        self._cache = OOBTree()

    def __repr__(self):
        name = self.__class__.__name__
        name = "%s '%s'" %(name, self.id)
        return "%s %s :: %s" %(name, self.field, [x for x in self._cache.items()])

    def get(self, key, default=None):
        return self._cache.get(key, default)

    def set(self, key, value):
        self._cache[key] = value
        self._p_changed

    def getCache(self, key):
        subcache = self.field.get(key, _marker)
        if subcache is _marker:
            cache = Cache(parent=self, id_=self.id)
            self.field[key] = cache
            subcache = self.field[key]
            self._p_changed
        return subcache
        
    def remove(self, key):
        val = self._cache.get(key)
        if val:
            del self._cache[key]
            self._p_changed=True
            for name, fcache in self.field.items():
                for slug, uid in fcache.items():
                    if uid==key:
                        del fcache[slug]

           
class Cache(PersistentMapping):

    def __init__(self, id_=None, parent=None):
        self.parent = parent
        self.id = id_
        self._reverse={}
        super(Cache, self).__init__()

    def getRaw(self, key):
        return super(Cache, self).__getitem__(key)

    def parentGet(self, uid, default=None):
        return self.parent.get(str(uid), default)

    def get(self, key, default=None):
        uid = super(Cache, self).get(key, default)
        return self.parentGet(uid, default)

    def set(self, key, value):
        slug, uid = key
        self[slug] = uid
        self._reverse[uid] = slug
        self.parent.set(uid, value)
        self._p_changed
        return value
        
    def __getitem__(self, key):
        retval = self.parentGet(key)
        if retval: return retval
        return self.getRaw(key)

    def __repr__(self):
        rep = super(Cache, self).__repr__()
        name = self.__class__.__name__
        name = "%s '%s'" %(name, self.id)
        return "%s %s" %(name, rep)

    def __delitem__(self, key):
        """
        trickle up deletion
        """
        uid = self.getRaw(key)
        self.parent.remove(key)
        
        super(Cache, self).__delitem__(key)

# @@ migrate to new name        
CACHE_KEY = 'Products.wicked.lib.factories.ContentCacheManager'

class BaseCacheManager(object):
    """abstract base"""
    implements(ICacheManager)
    adapts(IWickedFilter, IAmWicked)
    
    def __init__(self, wicked, context):
        self.context = context
        self.name = wicked.section

    @clearbefore
    def setName(self, name):
        self.name = name

    @memoizedproperty
    def cache_store(self):
        raise NotImplementedError

    def _get_cache(self, name=None):
        if not name:
            name = self.name
        cache = self.cache_store.getCache(name)
        return cache

    @memoizedproperty
    def cache(self):
        return self._get_cache()
    
    def get(self, key, default=None):
        return self.cache.get(key, default)
        
    def set(self, key, value):
        self.cache.set(key, value)
        return value
        
    def unset(self, key, use_uid=False):
        val = None
        if use_uid:
            for tkey, uid in self.cache.items():
                if uid == key:
                    val = self.get(tkey)
                    del self.cache[tkey]

        if self.cache.has_key(key):
            val = self.get(key)
            del self.cache[key]

        return val
    
    def remove(self, uid):
        self.cache_store.remove(uid)

    def reset(self, uid, value):
        self.cache_store.set(uid, value)


class UtilityCacheManager(BaseCacheManager):

    @memoizedproperty
    def cache_store(self):
        return getUtility(IWickedCacheStore)

    def name(): 
        def _pset(self, name, value):
            self._name=(name, IUID(self.context))
        def _pget(self, name):
            return self._name
        return locals()
    name = property(name())


class ContentCacheManager(BaseCacheManager):

    @memoizedproperty
    def cache_store(self):
        ann = IAnnotations(self.context)
        cache_store = ann.get(CACHE_KEY)
        if not cache_store:
            cache_store = CacheStore(id_='/'.join(self.context.getPhysicalPath()))
            ann[CACHE_KEY] = cache_store
        return cache_store






from interfaces import ITestObject
from zope.interface import alsoProvides as mark

def dummy(kdict, name='dummy', iface=ITestObject, bases=(object,)):
    """ factory for dummies """
    obj = type(name, bases, kdict)()
    mark(obj, iface)
    return obj 

def getToolByName(context, tool_name):
    """
    acquisition faker
    """
    return getattr(context, tool_name)

portal_path = "/path_to/portal"

class portal_url(object):
    def __init__(self, portal_path=portal_path):
        self.portal_path=portal_path
        
    def getPortalPath(self):
        return self.portal_path

_marker = object()
class pdo(dict):
    """
    pseudo data object....masquerades as a brain
    """

    def __init__(self, **kwargs):
        self.update(kwargs)

    def __getattr__(self, key):
        ret = self.get(key, _marker)
        if not ret is _marker:
            return ret
        raise AttributeError

    def getPath(self):
        return portal_path + "/%s" %self.getId
    
    def getObject(self):
        self.UID = lambda : self.UID
        return self



from Missing import Value as MissingValue
from Products.Archetypes.interfaces import IReferenceable
from Products.CMFCore.utils import getToolByName
from wicked.interfaces import IWickedFilter, IAmWicked, IWickedTarget
from wicked.at import config
from wicked.interfaces import IATBacklinkManager, IUID
from wicked.normalize import titleToNormalizedId as normalize
from relation import Backlink
from zope.interface import alsoProvides as mark
from zope.interface import implements
from wicked.utils import memoizedproperty, memoize, match, packBrain, cleanUID
from zope.component import adapts


class ATBacklinkManager(object):
    #@@ needs refactoring
    implements(IATBacklinkManager)
    adapts(IWickedFilter, IReferenceable)
    
    relation = config.BACKLINK_RELATIONSHIP
    refKlass = Backlink

    packBrain = staticmethod(packBrain)
    
    def __init__(self, wfilter, context):
        self.wfilter = wfilter
        self.context = context
        self.cm = wfilter.cache
        self.resolver = wfilter.resolver
        self.getMatch = wfilter.getMatch
        ## ATism: remove ASAP
        self.refcat = getToolByName(self.context, 'reference_catalog')
        self.suid = IUID(self.context)

    def getLinks(self):
        """
        Returns dataobjects representing backlinks
        """
        refbrains = self.refcat._queryFor(relationship=self.relation,
                                 tid=self.suid, sid=None)
        if refbrains:
            uids = [brain.sourceUID for brain in refbrains]
            ## XXX non-orthogonal
            return self.resolver.queryUIDs(uids)
        return []
    
    def _preplinks(self, links=dict()):
        return links and dict([(normalize(link), link) for link in links]) \
                     or dict()

    def manageLinks(self, new_links):
        # this has been heavily optimized
        # @@ but asyncing backlinking would help
                
        scope=self.wfilter.scope
        dups = set(self.removeLinks(new_links))

        resolver = self.resolver
        norm=tuple()
        for link in new_links:
            normalled=normalize(link)
            norm+=normalled,
            self.resolver.aggregate(link, normalled, scope)

        for link, normalled in zip(new_links, norm):
            match = self.getMatch(link, resolver.agg_brains, normalled=normalled)            
            if not match:
                match = self.getMatch(link, resolver.agg_scoped_brains, normalled=normalled)
            if not match:
                continue
            if match.UID != MissingValue and match.UID in dups:
                continue
            self.manageLink(match, normalled)

    def manageLink(self, obj, normalled):
        # need IObject iface for catalog brains
        if hasattr(obj, 'getObject'):
            # brain, other sort of pseudo object
            obj = obj.getObject()

        if not IReferenceable.providedBy(obj):
            # backlink not possible
            return

        mark(obj, IWickedTarget)
        self.refcat.addReference(obj,
                                 self.context,
                                 relationship=self.relation,
                                 referenceClass=self.refKlass)
        objuid = IUID(obj)
        path = '/'.join(obj.getPhysicalPath())
        data = dict(path=path,
                    icon=obj.getIcon(),
                    uid=objuid)
        
        self.cm.set((intern(normalled), objuid), [data])  

    def removeLinks(self, exclude=tuple()):
        oldlinks = self.getLinks()
        if not oldlinks:
            return set()
        
        exclude = self._preplinks(exclude)        
        dups = set([brain for brain in oldlinks if \
                    self.match(brain, exclude.get)])
        
        [self.remove(brain) for brain in set(oldlinks) - dups]
        return [self.cleanUID(b) for b in dups]

    cleanUID = staticmethod(cleanUID)

    def match(self, brain, getlink):
        """
        mmmm....turtle.
        @@ make efficient
        """
        link = getlink(brain.getId,
                       getlink(normalize(brain.Title), None))
        if link:
            return True
                
    def remove(self, brain):
        #XXX needs test
        objs = self.refcat._resolveBrains(\
            self.refcat._queryFor(self.suid, brain.UID, self.relation))
        for obj in objs:
            self.refcat._deleteReference(obj)
            self.cm.unset(obj.targetUID, use_uid=True)

    def unlink(self, uid):
        self.cm.remove(uid)



from Products.Archetypes.utils import shasattr
from Products.CMFCore.utils import getToolByName
from Testing import ZopeTestCase
from sets import Set
from wicked import utils
from wicked.interfaces import IWickedFilter, ICacheManager, IUID
from wickedtestcase import WickedTestCase
import os, sys, time
import traceback
import unittest
import pdb; st = pdb.set_trace

MARKER = dict(path='/apath',
              icon='anicon.ico',
              uid='uid')

class TestLinkCache(WickedTestCase):
    wicked_type = 'IronicWiki'
    wicked_field = 'body'

    def afterSetUp(self):
        """
        sets the body of page1 to be a wicked link of the id of page2
        """
        super(TestLinkCache, self).afterSetUp()
        field = self.page1.getField(self.wicked_field)
        field.set(self.page1, "((%s))" % self.page2.Title())
        self.field = field
        self.filter = utils.getWicked(field, self.page1)
        self.wicked_ccm = self.filter.cache
        self.wicked_ccm.name=field.getName()
        
    def test_linkGetsCached(self):
        field = self.field
        wicked_ccm = self.wicked_ccm
        pg2_id = self.page2.getId()
        val = wicked_ccm.get(pg2_id)
        self.failUnless(val)
        data=dict(path='/plone/Members/test_user_1_/dmv-computer-has-died',
                  icon='plone/document_icon.gif')
        data['uid']=IUID(self.page2)
        self.failUnlessEqual(set(val[0].items()), set(data.items()))

    def test_cacheIsUsed(self):
        field = self.field
        wicked_ccm = self.wicked_ccm
        pg2_id = self.page2.getId()
        wicked_ccm.set((pg2_id, IUID(self.page2)), [MARKER])
        value = self.getRenderedWickedField(self.page1)
        self.failUnless(MARKER['path'] in value)
        self.failIfWickedLink(self.page1, self.page2)
        

def test_suite():
    suite = unittest.TestSuite()
    from Testing.ZopeTestCase import ZopeDocTestSuite
    suite.addTest(unittest.makeSuite(TestLinkCache))
    return suite




from zope.interface import implementer
from zope.component import adapter
from Products.Archetypes.interfaces import IBaseObject
from wicked.interfaces import IUID


def initialize(context):
    """init for wicked.at"""
    pass


@implementer(IUID)
@adapter(IBaseObject)
def at_uid(atobj):
    return atobj.UID()



# we want to test each configuration to make sure what we think is
# happening does.
import unittest
from Products.PloneTestCase import ptc, five
from Products.PloneTestCase.layer import ZCML, PloneSite
from wicked.at.tests import wickedtestcase as wtc
from wicked.at.tests import test_cache, test_linking
from wicked.at.tests.wickedtestcase import WickedSite
from wicked.fieldevent import meta 
from Products.Five import zcml
import wicked.plone as here
from Products.PloneTestCase.layer import PloneSite 

from Testing import ZopeTestCase

# oh ghost of Tiran forgive me, but type substitution is important
test_klasses = (test_cache.TestLinkCache,
                test_linking.TestDocCreation,
                test_linking.TestWikiLinking,
                test_linking.TestLinkNormalization,
                test_linking.TestRemoteLinking)

ptc.setupPloneSite()

# this manhandles the zcml a bit
    

from registration import basic_type_regs


class SelectiveATCT(PloneSite):

    @classmethod
    def setUp(cls):
        app = ZopeTestCase.app()
        plone = app.plone
        for factory in basic_type_regs:
            factory(plone).handle()
        import transaction as txn
        txn.commit()
        ZopeTestCase.close(app)

    @classmethod
    def tearDown(cls):
        """thanks to the zcml load, teardown is not an option"""
        import transaction as txn
        app = ZopeTestCase.app()
        plone = app.plone
        for factory in basic_type_regs:
            factory(plone).handle(unregister=True)
        txn.commit()
        ZopeTestCase.close(app)


def make_wicked_suite(klasses, layer, content_type="Document", field="text"):
    """a factory to spitting out a new test class and new suite for
    these functional tests"""
    suite = unittest.TestSuite()

    for klass in klasses:
        newname = '%s_%s' %(klass.__name__, layer.__name__)
        bases = (klass,)
        newdict = dict(wicked_type=content_type,
                       wicked_field=field,
                       layer=layer,
                       setter='setText')
        newklass = type(newname, bases, newdict)
        new = unittest.makeSuite(newklass)
        suite.addTest(new)
        
    return suite

        
def test_suite():
    # this is overkill but proves the configs work
    suites = make_wicked_suite(test_klasses, SelectiveATCT) 
    return unittest.TestSuite(suites)
    
    



from zope.event import notify
from wicked.fieldevent.interfaces import IFieldEvent, IFieldRenderEvent, IFieldStorageEvent
from wicked.fieldevent.interfaces import IFieldValue, IFieldValueSetter, IField
from wicked.fieldevent.interfaces import FieldRenderEvent, FieldStorageEvent
from zope.component import subscribers, getMultiAdapter, handle
from zope.component import adapter, adapts
from zope.interface import implementer, implements

@adapter(IFieldEvent)
def notifyFieldEvent(event):
    field = event.field
    if IFieldRenderEvent.providedBy(event):
        event.value = getMultiAdapter((field, event), IFieldValue)

    handle(event.field, event.instance, event)
    
    if IFieldStorageEvent.providedBy(event):
        for ignore in subscribers((field, event), IFieldValueSetter):
            pass

def render(field_, instance, **kwargs):
    renderer=FieldRenderEvent(field_, instance, **kwargs)
    notify(renderer)
    return renderer.value

def store(field_, instance, value, **kwargs):
    store = FieldStorageEvent(field_, instance, value, **kwargs)
    notify(store)

# @@ separate IATField interfaces?
@implementer(IFieldValue)
@adapter(IField, IFieldRenderEvent)
def at_field_retrieval(field, event):
    return field.get(event.instance, **event.kwargs)

# @@ separate IATField interfaces?
@implementer(IFieldValueSetter)
@adapter(IField, IFieldStorageEvent)
def at_field_storage(field, event):
    field.set(event.instance, event.value, **event.kwargs)

## testing! ##

def test_suite():
    import unittest
    from zope.testing import doctest
    from zope.component import provideAdapter, provideHandler
    from zope.component import provideSubscriptionAdapter
    from zope.testing.cleanup import cleanUp
    
    def setUp(tc):
        # clean slate!
        cleanUp()

        # init event system
        from zope.component import event
        
        # register components
        provideHandler(notifyFieldEvent)
        provideAdapter(at_field_retrieval)
        provideSubscriptionAdapter(at_field_storage)

    def tearDown(tc):
        cleanUp()
        
    globs=globals()
    globs.update(locals())
    
    optflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
    rm = doctest.DocFileTest("README.txt",
                             package="wicked.fieldevent",
                             globs=globs,
                             setUp=setUp,
                             tearDown=tearDown,
                             optionflags=optflags)
    
    return unittest.TestSuite((rm,))



import os, sys, time
import unittest
from sets import Set
import traceback
from Testing import ZopeTestCase
from wickedtestcase import WickedTestCase, makeContent, TITLE2
from wicked.normalize import titleToNormalizedId
from wicked.at.config import BACKLINK_RELATIONSHIP
from Products.CMFCore.utils import getToolByName

class Base(WickedTestCase):
    wicked_type = 'IronicWiki'
    wicked_field = 'body'
    
    def replaceCreatedIndex(self):
        """ replace the 'created' index w/ a field index b/c we need
        better than 1 minute resolution for our testing """
        cat = getToolByName(self.portal, 'portal_catalog')
        cat.delIndex('created')
        cat.manage_addIndex('created', 'FieldIndex',
                            extra={'indexed_attrs':'created'})
        cat.manage_reindexIndex(ids=['created'])
    
    def demoCreate(self, **kw):
        self.login('test_user_1_')
        addview = self.page1.restrictedTraverse('@@wickedadd')
        addview.add_content(title=self.title, section=self.wicked_field, type_name='IronicWiki')

    def moveContent(self, obj, target):
        cps = obj.aq_parent.manage_copyObjects([obj.getId()])
        target.manage_pasteObjects(cps)



class TestWikiLinking(Base):
    
    def afterSetUp(self):
        super(TestWikiLinking, self).afterSetUp()
        self.set_text(self.page1, '((%s))' % TITLE2)

    def test_backlink(self):
        assert self.page1 in self.page2.getRefs(relationship=BACKLINK_RELATIONSHIP)

    def testforlink(self):
        self.failUnlessWickedLink(self.page1, self.page2)

    def testformultiplelinks(self):
        self.set_text(self.page1, '((DMV Computer has died))  ((Make another link))')
        self.failUnlessAddLink(self.page1)
                        
        self.failUnlessWickedLink(self.page1, self.page2)
        
    def testInexactTitleNotMatch(self):
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title With Extra')
        self.set_text(self.page1, "((W1 Title))")
        self.failIfWickedLink(self.page1, w1)
        self.failUnlessAddLink(self.page1)

    def testInexactTitleNotBlockLocalId(self):
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(self.folder, 'w2', self.wicked_type,
                         title='%s With Extra' % w1.id)
        self.set_text(self.page1, "((%s))" % w1.id)
        self.failUnlessWickedLink(self.page1, w1)
        self.failIfWickedLink(self.page1, w2)

    def testInexactLocalTitleNotBlockLocalTitle(self):
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(self.folder, 'w2', self.wicked_type,
                         title='%s With Extra' % w1.Title())
        self.set_text(self.page1, "((%s))" % w1.Title())
        self.failUnlessWickedLink(self.page1, w1)
        self.failIfWickedLink(self.page1, w2)

    def testInexactRemoteTitleNotBlockRemoteId(self):
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(self.folder, 'w2', self.wicked_type,
                         title='%s With Extra' % w1.id)
        w3 = makeContent(f2, 'w3', self.wicked_type,
                         title='W3 Title')
        self.set_text(w3, "((%s))" % w1.id)
        self.failUnlessWickedLink(w3, w1)
        self.failIfWickedLink(w3, w2)

    def testInexactRemoteTitleNotBlockRemoteTitle(self):
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(self.folder, 'w2', self.wicked_type,
                         title='%s With Extra' % w1.Title())
        w3 = makeContent(f2, 'w3', self.wicked_type,
                         title='W3 Title')
        self.set_text(w3, "((%s))" % w1.Title())
        self.failUnlessWickedLink(w3, w1)
        self.failIfWickedLink(w3, w2)

    def testDupLocalTitleMatchesOldest(self):
        self.replaceCreatedIndex()
        title = 'Duplicate Title'
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title=title)
        w2 = makeContent(self.folder, 'w2', self.wicked_type,
                         title=title)
        w3 = makeContent(self.folder, 'w3', self.wicked_type,
                         title=title)
        self.set_text(self.page1, "((%s))" % title)
        self.failUnlessWickedLink(self.page1, w1)

        # this also tests that deleting uncaches
        self.folder.manage_delObjects(ids=[w1.getId()])

        self.failUnlessWickedLink(self.page1, w2)

    def testDupRemoteIdMatchesOldest(self):
        self.replaceCreatedIndex()
        id = 'duplicate_id'
        f2 = makeContent(self.folder, 'f2', 'Folder')
        f3 = makeContent(self.folder, 'f3', 'Folder')
        f4 = makeContent(self.folder, 'f4', 'Folder')
        w1 = makeContent(f2, id, self.wicked_type,
                         title='W1 Title')
        # mix up the order, just to make sure
        w3 = makeContent(f4, id, self.wicked_type,
                         title='W3 Title')
        w2 = makeContent(f3, id, self.wicked_type,
                         title='W2 Title')
        self.set_text(self.page1, "((%s))" % id)
        self.failIfWickedLink(self.page1, w2)
        self.failIfWickedLink(self.page1, w3)
        self.failUnlessWickedLink(self.page1, w1)

        f2.manage_delObjects(ids=[w1.id])
        self.failIfWickedLink(self.page1, w2)

        # fails due to caching
        self.failUnlessWickedLink(self.page1, w3)

    def makeFolders(self, *args):
        folders = list()
        for id in args:
            folders.append(makeContent(self.folder, id, 'Folder'))
        return tuple(folders)
            
    def testDupRemoteTitleMatchesOldest(self):
        self.replaceCreatedIndex()
        title = 'Duplicate Title'
        
        f2, f3, f4 = self.makeFolders('f2', 'f3', 'f4')

        w1 = makeContent(f2, 'w1', self.wicked_type,
                         title=title)
        # mix up the order, just to make sure
        w3 = makeContent(f4, 'w3', self.wicked_type,
                         title=title)
        w2 = makeContent(f3, 'w2', self.wicked_type,
                         title=title)
        self.set_text(self.page1, "((%s))" % title)
        self.failUnlessWickedLink(self.page1, w1)
        self.failIfWickedLink(self.page1, w2)
        self.failIfWickedLink(self.page1, w3)

        f2.manage_delObjects(ids=[w1.id])
        self.failUnlessWickedLink(self.page1, w3)
        self.failIfWickedLink(self.page1, w2)

    def testLinkPersistsThroughMove(self):
        title = 'Move Me'
        f2, f3 = self.makeFolders('f2', 'f3')
        w1 = makeContent(f2, 'w1', self.wicked_type,
                         title=title)
        
        # check implicit resolution
        # this is a pre-test
        # should set cache
        self.set_text(self.page1, "((%s))" % title)
        self.failUnlessWickedLink(self.page1, w1)

        # move w1
        self.moveContent(w1, f3)

        w1.setTitle('new title to make sure we do not accidentally resolve')
        
        # check link did not change
        self.failUnlessWickedLink(self.page1, w1)

class TestDocCreation(Base):
    
    def afterSetUp(self):
        WickedTestCase.afterSetUp(self)
        self.title = 'Create a New Document'
        self.demoCreate()
        self.set_text(self.page1, '((%s))' %self.title)
        
    def testDocAdded(self):
        self.failUnless(getattr(self.folder,
                                titleToNormalizedId(self.title), None))

    def testBacklinks(self):
        newdoc = getattr(self.folder, titleToNormalizedId(self.title))
        backlinks = newdoc.getRefs(relationship=BACKLINK_RELATIONSHIP)
        self.failUnless(self.page1 in backlinks)
        
class TestLinkNormalization(Base):
    title = 'the monkey flies at dawn'

    def afterSetUp(self):
        super(TestLinkNormalization, self).afterSetUp()
        title1 = self.title 
        self.login('test_user_1_')
        self.newpage = self.clickCreate(self.page1, self.title)

    def clickCreate(self, page, title):
        """
        simulates browser interaction
        """
        addview = page.restrictedTraverse('@@wickedadd')
        addview.add_content(title=title, section=self.wicked_field, type_name='IronicWiki')
        self.set_text(page, "((%s))" %title ) #wha?
        return getattr(self.folder, titleToNormalizedId(title))

# @@ demonstrates issue with ids that are not tightly coupled to Title
# as of 1.0.1 still fails
##     def test_oldTitleWinsNewId(self):
##         # this will should fail
##         # if title changes don't trigger
##         # id changes
##         self.replaceCreatedIndex()
##         newtitle = 'I changed my mind'
##         self.page2.update(**dict(title=self.title))
##         self.newpage.update(**dict(title=newtitle))

##         # page one should still link to new page
##         # even though page2 has same title as link
##         self.failUnlessWickedLink(self.page1, self.newpage)

##         # delete newpage and recreate
##         # older title should beat newer id
##         self.loginAsPortalOwner()
##         self.folder.manage_delObjects([self.newpage.getId()])
##         self.newpage = self.clickCreate(self.page2, self.title)

##         self.failIfWickedLink(self.page1, self.newpage)
##         self.failUnlessWickedLink(self.page1, self.page2)

        
    def test_create_titlechange(self):
        # add content from link
        # test link
        # change title
        # test link
        title1 = self.title 

        # if this fails, wicked is not working period
        self.failUnlessWickedLink(self.page1, self.newpage)

        self.newpage.update(**dict(title='I changed my mind'))
        self.failUnlessWickedLink(self.page1, self.newpage)

class TestRemoteLinking(Base):
    
    def afterSetUp(self):
        super(TestRemoteLinking, self).afterSetUp()
        self.set_text(self.page1, '((%s))' % TITLE2)

    def testLocalIdBeatsRemoteId(self):
        self.replaceCreatedIndex()
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(f2, self.page1.getId(), self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(f2, 'w2_id', self.wicked_type,
                         title='W2 Title')
        self.set_text(w2, "((%s))" % self.page1.getId())
        self.failIfWickedLink(w2, self.page1)
        self.failUnlessWickedLink(w2, w1)

    def testLocalTitleBeatsRemoteId(self):
        self.replaceCreatedIndex()
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(f2, 'w1_id', self.wicked_type,
                         title=self.page1.id)
        w2 = makeContent(f2, 'w2_id', self.wicked_type,
                         title='W2 Title')
        self.set_text(w2, "((%s))" % self.page1.id)
        self.failUnlessWickedLink(w2, w1)
        self.failIfWickedLink(w2, self.page1)

    def testInexactLocalTitleNotBlockRemoteTitle(self):
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(f2, 'w2', self.wicked_type,
                         title='%s With Extra' % w1.Title())
        w3 = makeContent(f2, 'w3', self.wicked_type,
                         title='W3 Title')
        self.set_text(w3, "((%s))" % w1.Title())
        self.failUnlessWickedLink(w3, w1)
        self.failIfWickedLink(w3, w2)

    def testInexactLocalTitleNotBlockRemoteId(self):
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(self.folder, 'w1', self.wicked_type,
                         title='W1 Title')
        w2 = makeContent(f2, 'w2', self.wicked_type,
                         title='%s With Extra' % w1.id)
        w3 = makeContent(f2, 'w3', self.wicked_type,
                         title='W3 Title')
        self.set_text(w3, "((%s))" % w1.id)
        self.failUnlessWickedLink(w3, w1)
        self.failIfWickedLink(w3, w2)
        

def test_suite():
    from wicked.at.link import test_suite as btests
    suites = [unittest.makeSuite(tc) for tc in TestDocCreation, TestWikiLinking, TestLinkNormalization, TestRemoteLinking]
    suites.append(btests())
    suite = unittest.TestSuite(suites)
    return suite




import os
from cStringIO import StringIO
from Testing import ZopeTestCase
from Testing.ZopeTestCase import PortalTestCase
from Products.CMFCore.utils  import getToolByName
from Products.PloneTestCase import ptc
from Products.PloneTestCase.layer import PloneSite 
from wicked.atcontent.Extensions.Install import install as installWicked
import wicked.at.config as config
from wicked.testing.xml import xstrip as strip
from wicked.normalize import titleToNormalizedId
from wicked.registration import BasePloneWickedRegistration 
from zope.interface import Interface


ptc.setupPloneSite(products=['wicked.atcontent'])

TITLE1 = "Cop Shop"
TITLE2 = 'DMV Computer has died'

USELXML = False

import transaction as txn
#from collective.testing.debug import autopsy

class WickedSite(PloneSite):

    @classmethod
    def setUp(cls):
        app = ZopeTestCase.app()
        plone = app.plone
        reg = BasePloneWickedRegistration(plone)
        reg.handle()
        txn.commit()
        ZopeTestCase.close(app)

    @classmethod
    def tearDown(cls):
        app = ZopeTestCase.app()
        plone = app.plone
        reg = BasePloneWickedRegistration(plone)
        reg.handle(unregister=True)
        txn.commit()
        ZopeTestCase.close(app)

# This is the test case. You will have to add test_<methods> to your
# class in order to assert things about your Product.
class WickedTestCase(ptc.PloneTestCase):

    layer = WickedSite
    setter = 'setBody'

    def set_text(self, content, text, **kw):
        setter = getattr(content, self.setter)
        setter(text, **kw)
    
    def afterSetUp(self):
        # add some pages
        self.page1 = makeContent(self.folder,
                                 titleToNormalizedId(TITLE1),
                                 self.wicked_type, title=TITLE1)
        self.page2 = makeContent(self.folder,
                                 titleToNormalizedId(TITLE2),
                                 self.wicked_type, title=TITLE2)

    strip = staticmethod(strip)
 
    def getRenderedWickedField(self, doc):
        fieldname = self.wicked_field
        text = doc.getField(fieldname).get(doc)
        return self.strip(text)

    def failIfAddLink(self, doc):
        """ does wicked field text contain a wicked-generated add link? """
        # XXX make test stronger, support looking for specific links
        home_url= doc.absolute_url()
        text = self.getRenderedWickedField(doc)
        if home_url in text:
            self.fail("%s FOUND:\n\n %s" %(home_url, text))

    def failUnlessAddLink(self, doc):
        """ does wicked field text contain a wicked-generated add link? """
        # XXX make test stronger, support looking for specific links
        home_url= doc.absolute_url()
        text = self.getRenderedWickedField(doc)
        if not home_url in text:
            self.fail("%s NOT FOUND:\n\n %s" %(home_url, text))

    def failIfWickedLink(self, doc, dest):
        dest = dest.absolute_url()
        text = self.getRenderedWickedField(doc)
        if dest in text:
            self.fail("%s FOUND:\n\n %s" %(dest, text))

    failIfMatch = failIfWickedLink

    def failUnlessWickedLink(self, doc, dest):
        dest = dest.absolute_url()
        text = self.getRenderedWickedField(doc)
        if not dest in text:
            self.fail("%s NOT FOUND:\n\n %s" %(dest, text))

    failUnlessMatch = failUnlessWickedLink

    def hasAddLink(self, doc):
        """ does wicked field text contain a wicked-generated add link? """
        # XXX make test stronger, support looking for specific links
        return doc.absolute_url() in self.getRenderedWickedField(doc)

    def hasWickedLink(self, doc, dest):
        """ does wicked field text contain a resolved wicked link to
        the specified dest object?  """
        # XXX make test stronger
        return dest.absolute_url() in self.getRenderedWickedField(doc)



def makeContent(container, id, portal_type, **kw):
    container.invokeFactory(id=id, type_name=portal_type, **kw)
    o = getattr(container, id)
    return o



from Products.ATContentTypes.migration.walker import CatalogWalker
from wicked.at.migration.migrator import WickedDocMigrator
from Products.CMFCore.utils import getToolByName

def migrate_atdoc_wickeddoc(self):
    catalog = getToolByName(self, 'portal_catalog')
    portal = getToolByName(self, 'portal_url').getPortalObject()
    out = []

    migrator = WickedDocMigrator
    out.append('*** Migrating %s to %s ***\n' % (migrator.src_portal_type,
                                                 migrator.dst_portal_type))
    try:
        w = CatalogWalker(migrator, portal) # ATCT-0.2
    except AttributeError:
        w = CatalogWalker(portal, migrator) # ATCT-1.0
    w_result = w.go()
    if type(w_result) == type(''):
        out.append(w_result) # ATCT-0.2
    else:
        out.append('%s Migrated\n' % migrator.src_portal_type)

    wf = getToolByName(self, 'portal_workflow')
    count = wf.updateRoleMappings()
    out.append('Workflow: %d object(s) updated.' % count)

    catalog.refreshCatalog(clear=1)
    out.append('Portal catalog updated.')

    ttool = getToolByName(self, 'portal_types')
    doc_fti = ttool.getTypeInfo('Document')
    if doc_fti.Metatype() != 'WickedDoc':
        atct_tool = getToolByName(self, 'portal_atct')
        get_transaction().commit(1)
        atct_tool._changePortalTypeName('Document', 'ATDocument',
                                        global_allow=0,
                                        title='AT Document')
        get_transaction().commit(1)
        atct_tool._changePortalTypeName('WickedDoc', 'Document',
                                        global_allow=1,
                                        title='Page')
        out.append('Document types switched')

    return '\n'.join(out)



from zope.interface import implements, Interface, Attribute
from zope.interface.common.sequence import IReadSequence

class IField(Interface):
    """A Field: a behavioral object associated to a context and a
    value"""


try:
    from Products.Archetypes.interfaces import IATField
    class IATField(IField, IATField):
        """ at specific field interface """
except ImportError:
    pass


class IFieldValue(Interface):
    """ a value derived from the context. could be anything """

class IFieldValueSetter(Interface):
    """ stores a value derived from the context. """

class IFieldEvent(Interface):
    """ something has happened involving a field """
    value = Attribute("the value to render forth")
    instance = Attribute("the context object")
    field = Attribute("the field object")
    kwargs = Attribute("AT compatibility cruft")

    def __init__(field_, instance, **kwargs):
        """everything needed to make AT rendering go"""

class IFieldRenderEvent(IFieldEvent):
    """ an container to fetch a value for a field """
    
class IFieldStorageEvent(IFieldEvent):
    """ an container to store a value for a field """

class ITxtFilter(Interface):
    """Abstract Base for filters.

    Filters happen during object access (rendering for example) and
    include the context the object is being run in.
    """
    name = Attribute('name for this filter')

    def __init__(field, context, event):
        """to initialize.
        """

    def __call__():
        """execute filter. must be a callable.
        """

class ITxtFilterList(IReadSequence):
    """list of txtfilter names to apply"""

## events and exceptions ##

class FieldEvent(object):
    implements(IFieldEvent)
    value=None    
    def __init__(self, field_, instance, **kwargs):
        self.field = field_
        self.instance = instance
        self.kwargs = kwargs
    
class FieldStorageEvent(FieldEvent):
    implements(IFieldStorageEvent)
    def __init__(self, field_, instance, value, **kwargs):
        super(FieldStorageEvent, self).__init__(field_, instance, **kwargs)
        self.value = value

class FieldRenderEvent(FieldEvent):
    implements(IFieldRenderEvent)

class EndFiltrationException(Exception):
    """raise to stop continuation of txtfiltering"""



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from zope.interface import Interface

class ITestObject(Interface):
    """
    test iface marker
    """



import os, sys, time
import unittest
from sets import Set
import traceback

from Testing import ZopeTestCase
from wickedtestcase import WickedTestCase, makeContent
from wicked.normalize import titleToNormalizedId
from wicked.config import BACKLINK_RELATIONSHIP

stx_body = """
Here is some structured text data:

- with

- a

- bulleted

- list
"""

class TestWickedRendering(WickedTestCase):
    # an untest? 
    wicked_type = 'IronicWiki'
    wicked_field = 'body'
    
    def test_stxRendering(self):
        self.set_text(self.page1, stx_body, mimetype='text/structured')
        self.failUnless('<ul>' in self.page1.getBody())

def test_suite():
    from wicked.at.link import test_suite as doctests
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestWickedRendering))
    suite.addTest(doctests())
    return suite




import os, sys, time
import unittest
from sets import Set
import traceback

from Testing import ZopeTestCase
from wickedtestcase import WickedTestCase, makeContent
from wicked.normalize import titleToNormalizedId
from wicked.config import BACKLINK_RELATIONSHIP

has_atct = True
try:
    from Products import ATContentTypes
except ImportError:
    print "WARNING: ATContentTypes not installed, WickedDoc tests not running"
    has_atct = False

class TestWickedDoc(WickedTestCase):
    wicked_type = 'WickedDoc'
    wicked_field = 'text'

    def test_filterApplied(self):
        wd1 = makeContent(self.folder, 'wd1', 'WickedDoc',
                              title='WD1 Title')
        wd1.setText("((%s)) ((%s))" % (self.page1.Title(),
                                       "Nonexistent Title"))
        self.failUnlessWickedLink(wd1, self.page1)
        self.failUnlessAddLink(wd1)

def test_suite():
    suite = unittest.TestSuite()
    if has_atct:
        suite.addTest(unittest.makeSuite(TestWickedDoc))
    return suite





try:
    from Products.ATContentTypes.migration.migrator import CMFItemMigrator
    from Products.ATContentTypes.migration.common import registerMigrator
    from Products.ATContentTypes.migration.walker import CatalogWalker
except ImportError:
    CMFItemMigrator = None

if CMFItemMigrator != None:
    class WickedDocMigrator(CMFItemMigrator):
        walker = CatalogWalker
        map = {'getText': 'setText'}
        
        src_portal_type = 'Document'
        src_meta_type = 'ATDocument'
        dst_portal_type = 'WickedDoc'
        dst_meta_type = 'WickedDoc'

        def custom(self):
            self.new.setContentType(self.old.getContentType())

    registerMigrator(WickedDocMigrator)



PROJECTNAME            = "wicked"
SKINS_DIR              = 'skins'
GLOBALS                = globals()
BACKLINK_RELATIONSHIP  = 'Backlink->Source Doc'
FILTER_NAME = 'Wicked Filter'

try:
    # When Reference are in CMFCore
    from Products.CMFCore.reference_config import *
    from Products.CMFCore.ReferenceCatalog import Reference
    from Products.CMFCore.Referenceable import Referenceable
except ImportError:
    # And while they still live in Archetypes
    from Products.Archetypes.ReferenceEngine import Reference
    from Products.Archetypes.Referenceable import Referenceable
    from Products.Archetypes.config import REFERENCE_CATALOG as REFERENCE_MANAGER
    from Products.Archetypes.config import UID_CATALOG as UID_MANAGER



from wicked.registration import SelectiveRegistration
from Products.ATContentTypes.interface import IATNewsItem, IATEvent, IATDocument

def document_reg(site):
    return SelectiveRegistration(site, content=IATDocument)
document_reg.interface = IATDocument
document_reg.type = "Page"

def newsitem_reg(site):
    return SelectiveRegistration(site, content=IATNewsItem)
newsitem_reg.interface = IATNewsItem
newsitem_reg.type = "NewsItem"

def event_reg(site):
    return SelectiveRegistration(site, content=IATEvent)
event_reg.interface = IATEvent
event_reg.type = "Event"

basic_type_regs = document_reg, event_reg, newsitem_reg



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from general import dummy
from wicked.factories import ContentCacheManager     

def fakecacheiface(cached):
    def call(*args):
        return args[0]
    def set(*args):
        return "<link cached>"
    methods = dict(__call__=call, set=set) 
    cacheman=type('dummy',
                  (dict,), methods)()
    cacheman.update(cached)
    return cacheman

class query(object):
    
    brains = ['We are brains!']
    
    def scopedSearch(self):
        if self.chunk != 'dud':
            return self.brains
        
    def configure(self, chunk, normalized, scope):
        self.chunk = chunk
        self.normalized = normalized
        self.scope = scope
        
    def search(self):
        if self.chunk != 'dud' and self.chunk != 'scoped':
            return self.brains

def argchug(rets):
    def function(*args, **kwargs):
        return rets
    
def fakefilter():
    def conf(*args, **kw):
        self = list(args).pop()
        [setattr(self, k, v) for k, v in kw.items()]
        return kw
    conf = classmethod(conf)
    kdict = dict(configure=conf, scope='/scope/')
    wfilter = dummy(kdict, name='wfilter')
    wfilter.query_iface = query()
    wfilter.getMatch = argchug(('uid', 'link'))
    wfilter.resolver = query()
    wfilter.section=hash(wfilter)
    wfilter.cache=ContentCacheManager(dummy(dict()))
    return wfilter



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit@kalistra.com>
#   - and contributors
#
##########################################################

"""
IronicWiki
~~~~~~~~~~

a demonstration of wicked's capabilities confined to the content type
wicked aspires to make obsolete

"""
from AccessControl import ClassSecurityInfo
from Products.Archetypes import public as atapi
from Products.CMFCore import permissions as CMFCorePermissions
from Products.CMFCore.utils import getToolByName
from wicked.at.field import WikiField
from wicked.atcontent import zope2
from wicked.interfaces import IAmWicked
from zope.annotation.interfaces import IAttributeAnnotatable 
from zope.interface import implements


class IronicWiki(atapi.BaseContent):
    """ Ironic wiki Portal Content """
    implements(IAmWicked, IAttributeAnnotatable)
    archetype_name = portal_type = meta_type = 'IronicWiki'
    schema = atapi.BaseSchema.copy() + atapi.Schema((
        WikiField( "body",
                   primary=True,
                   filters=('Wicked Filter',),
                   default_content_type='text/structured',
                   default_output_type='text/html',
                   allowable_content_types = ('text/structured',
                                              'text/restructured',
                                              'text/html',
                                              'text/plain',
                                              'text/plain-pre'),
               
                   widget=atapi.RichWidget( description = "The body text of the document.",
                                            description_msgid = "help_body_text",
                                            label = "Body text",
                                            label_msgid = "label_body_text",
                                            rows = 25,
                                            i18n_domain = "plone")),
        ))


atapi.registerType(IronicWiki, zope2.PROJECTNAME)



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from zope.interface import implements, Interface
from general import dummy

class Filter(object):
    """
    dummy
    """
    implements(Interface)
    
    def __init__(self, context):
        self.context = context

    def _filterCore(self, *args, **kwargs):
        pass

    def filter(self, *args, **kwargs):
        pass
    
    section='body'

portal_tools = dict()
def getToolByName(context, toolname, default):
    return portal_tool.get(toolname, default)

def backlinkTools():
    import Products.CMFCore.utils
    Products.CMFCore.utils.getToolByName = getToolByName
    from Products import AdvancedQuery
    from Products.CMFCore.CatalogTool import CatalogTool
    from Products.Archetypes.ReferenceEngine import ReferenceCatalog, UIDCatalog
    return dict(portal_workflow=None,
                portal_catalog=CatalogTool(),
                reference_catalog=ReferenceCatalog(id='portal_reference'),
                portal_uid=UIDCatalog(id='portal_uid'))

def setupTools(tools):
    portal_tools.update(tools)

from general import pdo

## class pRefCat(dict):
##     def _queryFor(self, **kw):
##         self.get(kw['tid']) 

##     def addReference(self, obj, target, *args, **kwargs):
##         attrs= dict(targetUID=target.UID())

##         if not self.has_key(self[obj.UID()]):
##             self[obj.UID()] = []
            
##         self[obj.UID()].append(pdo(attrs))

##     def _resolveBrains(self, suid, tuid, *args):
##         return [x for x in self.get(suid) if x.targetUID==tuid].pop()
    
##    def _deleteReference(self, obj):
##         del self[obj.UID()]



from wicked import utils
from wicked.interfaces import IWickedLink, IWickedEvent, WickedContentAddedEvent
from zope.component import handle, adapter
from zope.event import notify
from zope.interface import implements

_marker=object()

class WickedAdd(object):

    def __init__(self, context, request):
        self._context = (context,)
        self.request = request

    def notify_content_added(self, newcontent, title, section):
        notify(WickedContentAddedEvent(self.context, newcontent, title, section, self.request))
        
    @property
    def context(self):
        return self._context[0]

    def add_content(self, title=None, section=None, type_name=_marker):
        raise NotImplementedError

    def addMenu(self):
        return 1


_marker=object()

class BasicLink(object):
    implements(IWickedLink)
    section = _marker
    
    def __init__(self, context, request):
        self._context = (context,)
        self.request = request
        self.counter=utils.counter()
        
    @property
    def context(self):
        return self._context[0]

    @utils.memoizedproperty
    def howmany(self):
        return len(self.links)

    @utils.memoizedproperty
    def multiple(self):
        return self.howmany > 1

    @utils.memoizedproperty
    def links(self):
        """
        calculates urls

        abstract zope2ism
        """
        for link in self._links:
            link['url'] = self.request.physicalPathToURL(link['path'], 0)
        return self._links

    @property
    def singlelink(self):
        return self.links[0]

    @property
    def count(self):
        return self.counter.next()

    @utils.clearbefore
    def load(self, links, chunk):
        self._links = links
        self.chunk = chunk


@adapter(IWickedEvent)
def redispatch(event):
    handle(event.context, event)


def test_suite():
    import unittest
    from zope.testing import doctest
    suite = doctest.DocFileSuite('link.txt',
                                 globs=globals(),
                                 package='wicked',
                                 optionflags=doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS)
    return suite



if __name__=="__main__":
    unittest.TextTestRunner().run(test_suite())



import unittest
from zope.testing import doctest

def test_suite():
    from Testing.ZopeTestCase import ZopeDocFileSuite
    try:
        from Products.PloneTestCase.layer import ZCMLLayer
    except ImportError:
        from collective.testing.layer import ZCMLLayer
    suite = ZopeDocFileSuite('README.txt',
                             package="wicked.atcontent",
                             optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
                             )
    suite.layer = ZCMLLayer
    return unittest.TestSuite((suite))
    




##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from normalize import titleToNormalizedId as normalize
from zope.component import getMultiAdapter
from wicked.interfaces import IWickedFilter, WickedEvent

def linkcache(func):
    def cache(wfilter, chunk, normalized):
        # cache depends on query and match
        # this could use some untangling
        # generic function? 
        value = wfilter.cache.get(normalized)
        if not value:
            value = func(wfilter, chunk, normalized)
            if value:
                uid = value[0]['uid']
                wfilter.cache.set((normalized, uid), value)
        return value
    return cache

def getWicked(field, instance, event=None):
    if not event:
        event = WickedEvent()
    return getMultiAdapter((field, instance, event), IWickedFilter)


def match(query):
    def match(self, best_match=True):
        data = query(self)
        if data and best_match:
            return [getMatch(self.chunk, data, normalled=self.normalled)]
        return data
    return match


def cleanUID(brain):
    """takes a brain, returns a hashable value"""
    from Missing import Value
    uid = brain.UID
    if uid is Value:
        uid = brain.data_record_id_
    return uid


def packBrain(brain):
    """
    converts dataobjects in to template ready dictionaries

    >>> from Missing import Value
    >>> class brain(object):
    ...     def getPath(self):
    ...         return '/path'
    ...     data_record_id_=123
    ...     UID='unique'
    ...     getIcon='/donkey.gif'
    >>> packBrain(brain())['uid']
    'unique'

    packBrain needs to filter out Missing.Value that make creep in.

    >>> fbrain = brain()
    >>> brain.UID=Value
    >>> packBrain(fbrain)['uid']
    123
    """
    # missing values are not hashable nor can they be dict keys.  they
    # should never see the light of day. In the rare case that the UID
    # index is updated for an object, substitute the record id

    uid = cleanUID(brain)
    
    return dict(path=brain.getPath(),
                icon=brain.getIcon,
                uid=uid,
                rid=brain.data_record_id_)


def getMatch(chunk, brains, normalled=None):
    """
    Given a set of query results and the wicked link text, return
    the single brain that represents the correct object to link to, or
    None.
    
    Assumes that brains are already ordered oldest to newest, so
    the first absolute match is the one returned.  Matches on id
    take priority over matches on title

    all strings are normalized and interned for comparison matches.

    >>> from testing.general import pdo
    >>> mkbrain = lambda i: pdo(getId='-'.join([str(x) for x in i]), Title='%s %s' %i, created=i[1])
    >>> seed = zip('abc', range(3))
    >>> brains = [mkbrain(i) for i in seed]
    >>> chunk = ''
    >>> normalled = ''

    Test null conditions

    >>> getMatch(chunk, brains)
    >>> getMatch(chunk, brains, normalled)
    >>> getMatch(chunk, brains[:1], normalled)

    Test single brain matches

    >>> getMatch('', brains[:1], 'a-0').getId
    'a-0'
    
    >>> getMatch(brains[0].getId, brains[:1], 'blah').getId
    'a-0'

    >>> getMatch(brains[0].Title, brains[:1]).getId
    'a-0'

    Test multi brain matches. brain 0 should win over brain 3
    for all matches

    >>> from copy import copy
    >>> newbrain = copy(brains[0])
    >>> newbrain.update(dict(created=3))
    >>> brains =   brains + [newbrain]
    >>> getMatch('', brains, 'a-0').created
    0
    
    >>> getMatch(brains[0].getId, brains).created
    0

    >>> getMatch(brains[0].Title, brains).created
    0

    Test title to chunk match

    >>> brains[3].Title='A unique title'
    >>> getMatch(brains[3].Title, brains).Title
    'A unique title'
    """
    normalled_chunk = normalled
    if not normalled_chunk:
        normalled_chunk = normalize(chunk)
    normalled_chunk = intern(normalled_chunk)
    if not isinstance(brains, list):
        # make a copy to AdvancedQuery sequencing issues
        brains = [x for x in brains]

    # inspect single return case

    if len(brains) == 1 and \
           (intern(brains[0].getId) is normalled_chunk \
            or intern(brains[0].getId.strip()) is intern(chunk.strip()) \
            or intern(normalize(brains[0].Title)) is normalled_chunk):
        return brains[0]

    # first, match id

    # reversing the brains into a dict clobbers younger matches with
    # the same id. we'll match against the normalled chunk, then the
    # chunk (for single work chunks)

    btup = [(brain.getId, brain) for brain in brains]
    id_dict = dict(reversed(btup))
    for unk in normalled_chunk, chunk,:
        if id_dict.has_key(unk):
            return id_dict[unk]

    # second, match Title
    brains=[brain for brain in brains \
            if intern(normalize(brain.Title)) is normalled_chunk]
    
    return brains and brains[0] or None

def configure(self, **attrs):
    """
    For runtime configuration of filter.
    
    @param attrs: dict of 0-N key value pairs that
    will be set on the filter for the duration of
    its existence
    >>> from testing.general import portal_url, dummy, pdo
    >>> troll = dummy(dict(_configure_exclude=dict(fo=True)))
    >>> configure(troll, **dict(fi='fie', fo='fum'))
    {'fo': 'fum'}
    
    >>> getattr(troll, 'fi', 'No fie?!!')
    'fie'
    
    When no term are exclude, configure should
    set and remove all kwargs passed in
    
    >>> configure(troll, **dict(englishman='yum'))
    {}
    
    >>> getattr(troll, 'englishman')
    'yum'
    
    Make sure a dict gets returned even if
    kwargs are empty
    
    >>> configure(troll, **dict())
    {}
    """
    [(setattr(self, key, attrs[key]), attrs.__delitem__(key)) \
     for key in attrs.keys() \
     if not self._configure_exclude.has_key(key)]
    return attrs and attrs


def counter():
    count=0
    while True:
        count+=1
        yield count


_marker = object()
class Memoizer(object):
    propname = '_mp_cache'
    def clearbefore(self, func):
        def clear(*args, **kwargs):
            inst=args[0]
            if hasattr(inst, self.propname):
                delattr(inst, self.propname)
            return func(*args, **kwargs)
        return clear

    def memoizedproperty(self, func):
        return property(self.memoize(func))

    def memoize(self, func):
        def memogetter(*args, **kwargs):
            inst = args[0]
            cache = getattr(inst, self.propname, dict())
            key=hash((func.__name__, args, frozenset(kwargs)))
            val = cache.get(key, _marker)
            if val is _marker:
                val=func(*args, **kwargs)
                cache[key]=val
                setattr(inst, self.propname, cache)
            return val
        return memogetter


_m = Memoizer()
memoize = _m.memoize
memoizedproperty = _m.memoizedproperty
clearbefore = _m.clearbefore
    

def test_suite():
    import unittest
    from zope.testing import doctest
    optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
    return doctest.DocTestSuite('wicked.utils',
                                optionflags=optionflags)

if __name__=="__main__":
    unittest.TextTestRunner().run(test_suite())




from wicked.at.seeker import IWickedQuery, AdvQueryMatchingSeeker
from wicked.cache import ICacheManager, ContentCacheManager
from wicked.txtfilter import WickedFilter, IWickedFilter
from wicked.txtfilter import IAmWickedField, IFieldEvent
from wicked.txtfilter import BrackettedWickedFilter, BacklinkRegistrationProxy
from wicked import utils
from wicked.fieldevent.interfaces import IFieldRenderEvent, IFieldStorageEvent
from wicked.fieldevent.interfaces import IFieldValueSetter
from wicked.interfaces import IAmWicked
from wicked.interfaces import IAmWickedField
from wicked.interfaces import IWickedFilter
from wicked.interfaces import IWickedEvent

from wicked.txtfilter import BrackettedWickedFilter
from wicked.txtfilter import WickedFilter, wicked_listener
from zope.app.schema.vocabulary import IVocabularyFactory
from zope.component import getSiteManager
from zope.interface import implements
from zope.schema.vocabulary import SimpleVocabulary

try:
    import pkg_resources
    get_points = pkg_resources.iter_entry_points
except ImportError:
    get_points = lambda :[]


class BaseWickedRegistration(object):
    """abstract base: add a docstring to display a description on the
    wicked dashboard"""
    title = "((Basic))"
    field = IAmWickedField
    content = IAmWicked
    event = IFieldRenderEvent
    txtfilter = WickedFilter
    txtfilter_provides = IWickedFilter
    subscriber = (wicked_listener,)
    
    def __init__(self, context, **kw):
        self.site = context
        for key, item in kw.items():
            if hasattr(self, key):
                setattr(self, key, item)

    @utils.memoizedproperty
    def sm(self):
        return getSiteManager(self.site)

    @property
    def required(self):
        return (self.field, self.content, self.event)
    
    def handle(self, unregister=False):
        # @@ add logging
        handle_adapter_reg = self.sm.registerAdapter
        handle_subscriber_reg = self.sm.registerHandler
        handle_sub_adapter_reg = self.sm.registerSubscriptionAdapter
        if unregister:
            handle_adapter_reg = self.sm.unregisterAdapter
            handle_subscriber_reg = self.sm.unregisterHandler
            handle_sub_adapter_reg = self.sm.unregisterSubscriptionAdapter
        
        self.handle_txtfilter(handle_adapter_reg)
        handle_subscriber_reg(self.subscriber[0], required=self.required)
        handle_sub_adapter_reg(self.backlink_handler)
        return (handle_adapter_reg, handle_subscriber_reg, handle_sub_adapter_reg)

    def handle_txtfilter(self, handle, event=None, txtfilter=None, unregister=False):
        """helper for common reg of txtfilter"""
        txtfilter = txtfilter and txtfilter or self.txtfilter
        required = list(self.required)
        if event:
            required[2] = event

        handle(self.txtfilter,
               required=required,
               provided=self.txtfilter_provides)

    def is_installed(self):
        raise NotImplemented


class BasePloneWickedRegistration(BaseWickedRegistration):
    """Basic wicked behavior for Plone. Registers a fieldevent
    subscriber and the wicked filter."""
    backlink_handler = BacklinkRegistrationProxy


class BasePloneMediaWickedRegistration(BasePloneWickedRegistration):
    """Same as basic but with square bracket style linking"""
    title = "[[Bracketted]]"
    txtfilter = BrackettedWickedFilter

class SelectiveRegistration(BaseWickedRegistration):
    """by dropping direct registration to IAmWicked for content, we
    can selectively turn on and off the creation of wiki behavior by
    content interface. To do this though, we have to hand register a
    seeker and a cache manager"""
    cache = ContentCacheManager
    cache_provides = ICacheManager
    seeker = AdvQueryMatchingSeeker
    seeker_provides = IWickedQuery
    other_events = (IWickedEvent, IFieldStorageEvent)
    backlink_handler = BacklinkRegistrationProxy

    @property
    def cache_required(self):
        return (self.txtfilter_provides, self.content)
    
    def handle(self, unregister=False):
        handle_adapter_reg, handle_subscriber_reg, handle_sub_adapter_reg = super(SelectiveRegistration, self).handle(unregister=unregister)
        self.handle_cache(handle_adapter_reg)
        self.handle_seeker(handle_adapter_reg)
        
        for event in self.other_events:
            self.handle_txtfilter(handle_adapter_reg, event=event)
        return (handle_adapter_reg, handle_subscriber_reg)

    def handle_cache(self, handle):
        handle(self.cache,
               required=self.cache_required,
               provided=self.cache_provides)

    def handle_seeker(self, handle):
        handle(self.seeker,
               required=(self.content,),
               provided=self.seeker_provides)
    

## vocabulary

# @@ this will have to wait since plone does not have an egg story yet

class BaseConfigurationOptionsFactory(object):
    implements(IVocabularyFactory)

    def __call__(self, context):
        entry_points = get_points("wicked.base_registration")
        items = [(bc.name, bc.load().title) for bc in entry_points]
        items.sort()
        items.insert(0, ('off', 'Off'))
        return SimpleVocabulary.fromItems(items)

class CacheConfigurationOptionsFactory(object):
    implements(IVocabularyFactory)

    def __call__(self, context):
        return SimpleVocabulary.fromItems(list())


CacheConfigurationOptions = CacheConfigurationOptionsFactory()
BaseConfigurationOptions = BaseConfigurationOptionsFactory()




from wicked.fieldevent import render, store
from zope.interface import Interface, implementer
from zope.interface import noLongerProvides, alsoProvides
from zope.component import adapter
from zope.component.interface import provideInterface
from wicked.fieldevent.interfaces import IFieldValue
from wicked.fieldevent.interfaces import IFieldValueSetter
import itertools
import zope.component.zcml
import zope.configuration.fields as fields
import zope.interface
import zope.schema
import wicked.fieldevent
from zope.component import provideAdapter, provideHandler
from zope.component import provideSubscriptionAdapter


class NewLineTokens(fields.Tokens):
    """ token field that splits on newline not space """
    
    def fromUnicode(self, u):
        u = u.strip()
        if u:
            vt = self.value_type.bind(self.context)
            values = []
            for s in u.split('\\n'):
                try:
                    v = vt.fromUnicode(s)
                except schema.ValidationError, v:
                    raise InvalidToken("%s in %s" % (v, u))
                else:
                    values.append(v)
        else:
            values = []

        self.validate(values)
        return values


class IATFieldDecoratorDirective(Interface):
    fieldclass = fields.GlobalObject(
        title=u"Field Class",
        required=True
        )

class IATSchemaFieldsImplementDirective(Interface):
    atclass = fields.GlobalObject(
        title=u"Archetypes Schema Object",
        required=True
        )

    implements = fields.GlobalObject(
        title=u"Archetypes Schema Object",
        required=True
        )

    fields = NewLineTokens(
        title=u"Names of fields to be marked",
        description=u"""
        Each field enumerated here will now directly provide the
        interfaces in implements.
        """,
        required=True,
        value_type=zope.schema.TextLine(missing_value=str()))
    

def decorate_at_field(_context=None, fieldclass=None, zcml=True):
    if not fieldclass in _fieldclass_monkies:

        # register adapters to call original methods

        _get=fieldclass.get
        _set=fieldclass.set    
        
        @implementer(IFieldValue)
        @adapter(fieldclass, wicked.fieldevent.IFieldRenderEvent)
        def field_value(field, event):
            return _get(field, event.instance, **event.kwargs)

        @implementer(IFieldValueSetter)
        @adapter(fieldclass, wicked.fieldevent.IFieldStorageEvent)
        def field_value_set(field, event):
            _set(field, event.instance, event.value, **event.kwargs)

        if zcml:
            # do proper zopeish config
            zope.component.zcml.adapter(_context, (field_value,))
            zope.component.zcml.subscriber(_context,
                                           factory=field_value_set,
                                           provides=IFieldValueSetter)
            _context.action(
                discriminator = (fieldclass, monkey_getset),
                callable = monkey_getset,
                args = (fieldclass,)
                )
        else:
            # hook it up no delay
            monkey_getset(fieldclass)
            provideAdapter(field_value)
            provideSubscriptionAdapter(field_value_set)


_fieldclass_monkies=[]
def monkey_getset(fieldclass):
    global _fieldclass_monkies
    _get=fieldclass.get
    _set=fieldclass.set    
    fieldclass.set = store
    fieldclass.get = render
    fieldclass.__original_get=_get
    fieldclass.__original_set=_set
    _fieldclass_monkies.append(fieldclass)


def schemafields_implement(_context=None, atclass=None, implements=None, fields=[], zcml=True):
    if zcml:
        _context.action(
            discriminator = None,
            callable = provideInterface,
            args = ('', implements)
            )
    schemafields_provide(implements, atclass.schema, fields)        


_schemafields_marked=[]
def schemafields_provide(implements, schema, fieldnames, cleanup=False):
    action = noLongerProvides
    if not cleanup:
        global _schemafields_marked
        if (implements, schema, fieldnames) in _schemafields_marked:
            return
        _schemafields_marked.append((implements, schema, fieldnames))
        action = alsoProvides

    for name in fieldnames:
        field=schema[name]
        try:
            action(field, implements)
        except ValueError:
            pass


def cleanup_schemafields_provide():
    global _schemafields_marked
    for implements, schema, fieldnames in _schemafields_marked:
        schemafields_provide(implements, schema, fieldnames, cleanup=True)
    _schemafields_marked=[]


def cleanup_decorate_at_field():
    # unmonkey!
    global _fieldclass_monkies 
    for class_ in _fieldclass_monkies:
        class_.get = class_.__original_get
        class_.set = class_.__original_set
    _fieldclass_monkies = []


def cleanUp():
    cleanup_decorate_at_field()
    cleanup_schemafields_provide()    


from zope.testing.cleanup import addCleanUp
addCleanUp(cleanUp)
del addCleanUp


def test_suite():
    import unittest
    from zope.testing import doctest
    from zope.testing.cleanup import cleanUp

    def setUp(tc):
        # clean slate!
        cleanUp()
        provideHandler(wicked.fieldevent.notifyFieldEvent)

    def tearDown(tc):
        cleanUp()

    globs=globals()
    globs.update(locals())
    
    optflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
    meta = doctest.DocFileTest("meta.txt",
                             package="wicked.fieldevent",
                             globs=globs,
                             setUp=setUp,
                             tearDown=tearDown,
                             optionflags=optflags)
    
    return unittest.TestSuite((meta,))



# ganked from zopt and sfive
import os
from StringIO import StringIO
USELIBXML = False

slug = """ <div><some tag="true">
               <other />            </some>
     </div>
     """
xslt = """
<xsl:stylesheet version='1.0' xmlns:xsl='http://www.w3.org/1999/XSL/Transform'>
  <xsl:output method="html" indent="yes"/>
  <xsl:strip-space elements="*"/>
  <xsl:template match="@*|node()">
    <xsl:copy><xsl:apply-templates select="@*|node()"/></xsl:copy>
  </xsl:template>
</xsl:stylesheet>
"""

#import libxml2, libxslt
xsltfile = os.path.join(os.path.dirname(__file__), 'strip.xsl')
 	
#use stylesheet as global

try:
    import libxml2, libxslt
    styledoc=libxml2.parseFile(xsltfile)
    style=libxslt.parseStylesheetDoc(styledoc)

    def libstrip(text):
        """
        strip out whitespace
        >>> print xstrip(slug)
        <div><some tag="true"><other></other></some></div>
        ...
        """
        encoding = 'UTF-8'
        INDENT = True
        try:
            doc = libxml2.parseDoc(text)
            res = style.applyStylesheet(doc, None)
            # XXX: this raises an unhandled c exception
            #res = style.saveResultToString(resdoc)
            out = res.serialize(encoding=encoding, format=INDENT)
        finally:
            try: doc.freeDoc()
            except: pass
            try: res.freeDoc()
            except: pass
        return out

except ImportError:
    libstrip = None


def lxmlstrip(text):
    """
    strip out whitespace
    >>> print xstrip(slug)
    <div><some tag="true"><other></other></some></div>
    ...
    """
    if not text:
        return ''
    
    from lxml import etree
    xslt_doc = etree.XML(xslt)
    stripper = etree.XSLT(xslt_doc)
    doc = etree.XML(text)
    result = stripper(doc)
    val = str(result)
    return val

if libstrip and USELIBXML:
    xstrip = libstrip
else:
    xstrip = lxmlstrip

import unittest
from zope.testing import doctest
optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
def test_suite():

    return unittest.TestSuite((
        doctest.DocTestSuite('xml', optionflags=optionflags)
        ))

if __name__=="__main__":
    unittest.TextTestRunner().run(test_suite())



from Products.CMFCore.utils import getToolByName
from Products.Five.utilities.marker import mark
from wicked.at import config
from wicked.interfaces import IWickedBacklink 
from wicked.interfaces import IWickedTarget

def upgrade09(site):
    refcat = getToolByName(site, 'reference_catalog')
    refs = (brain.getObject() for brain in \
               refcat(relationship=config.BACKLINK_RELATIONSHIP))
    
    for ref in refs:
        if ref:
            obj=ref.getSourceObject()
            mark(obj, IWickedTarget)
            if not IWickedBacklink.providedBy(ref):
                print ref
                mark(ref, IWickedBacklink)



import pkg_resources

BASE_OPTIONS="wicked.base_registration"
CACHE_OPTIONS="wicked.cache_manager"

class WickedDashboard(object):
    """control panel for wicked manipulation"""

    def __init__(self, context, request):
        self.context = context
        self.request = request








##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit@kalistra.com>
#   - Rob Miller <rob@kalistra.com> (RaFromBRC)
#   - and contributors
#
##########################################################
from Products.Archetypes import public as atapi
from Products.Archetypes.Registry import registerField
from Products.Archetypes.interfaces import IBaseObject
from ZPublisher.HTTPRequest import FileUpload
from wicked import utils
from wicked.fieldevent.interfaces import IFieldRenderEvent, IFieldStorageEvent
from wicked.fieldevent.interfaces import ITxtFilterList, IFieldValueSetter
from wicked.interfaces import IAmWicked, IAmWickedField, IValueToString
from wicked.txtfilter import WickedFilter
from zope.component import adapter, adapts
from zope.interface import implements, implementer
from wicked.at.interfaces import IAmATWickedField
from cStringIO import StringIO

class WikiField(atapi.TextField):
    """ drop-in wiki """
    implements(IAmWickedField)

    _properties = atapi.TextField._properties.copy()
    _properties.update({
        'scope': '',
        })

registerField(WikiField,
              title='Wiki',
              description='Text field capable of wiki style behavior')


@implementer(IValueToString)
@adapter(atapi.BaseUnit, IAmATWickedField)
def baseunit_to_string(value, field):
    """this avoid a security proxy that will foul regexes"""
    return value.getRaw()


@implementer(IValueToString)
@adapter(FileUpload, IAmWickedField)
def fileupload_to_string(value, field):
    """a file was uploaded, get the (possibly transformed) value"""
    return field.get(instance, skip_filters=True)






def test_suite():
    import unittest, doctest
    from Testing.ZopeTestCase import ZopeDocFileSuite
    try:
        from Products.PloneTestCase.layer import ZCMLLayer
    except ImportError:
        from collective.testing.layer import ZCMLLayer
    from wicked.utils import test_suite as utilsuite
    from wicked.link import test_suite as linksuite

    cachemanager = ZopeDocFileSuite('cache.txt',
                                    package='wicked',
                                    optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
                                    )
    cachemanager.layer = ZCMLLayer
    return unittest.TestSuite((utilsuite(), linksuite(), cachemanager))



from wicked.fieldevent.interfaces import ITxtFilter, IField
from wicked.fieldevent.interfaces import ITxtFilterList, EndFiltrationException
from wicked import utils
from zope.component import queryMultiAdapter, subscribers
from zope.interface import implements
from interfaces import ITxtFilter


def txtfilter_output(field, instance, event):
    """a run once subscriber to process text in a pipeline"""
    
    if getattr(event, '_txtfiltered_', False):
        return
    
    filter_names = queryMultiAdapter((field, instance, event), ITxtFilterList)
    if not filter_names:
        return

    txts = subscribers((field, instance, event), ITxtFilter)
    txtmap = dict([(f.name, f) for f in txts])

    for name in filter_names:
        try:
            txtfilter = txtmap.get(name, None)
            if callable(txtfilter):
                txtfilter()
        except EndFiltrationException, e:
            break
        
    event._txtfiltered_=True


class TxtFilter(object):
    """Abstract Base for Filtration
    """
    implements(ITxtFilter)
    
    name = None    # required
    pattern = None
    
    def __init__(self, field, context, event):
        self.context = context
        self.field = field
        self.event = event

    @utils.memoizedproperty
    def chunks(self):
        """Simple text replacement via co-op with the modules"""
        return self.pattern.split(self.event.value)

    @utils.memoizedproperty
    def dynamic(self):
        """tricky ben saller split"""
        return self.chunks[1::2]

    @utils.memoizedproperty
    def filtered_chunks(self):
        return [self._filterCore(d, **self.event.kwargs) for d in self.dynamic]

    @property
    def filtered_text(self):
        """join the two lists (knowing that len(text) == subs+1)"""
        return ''.join(ijoin(self.chunks[::2], self.filtered_chunks))
    
    def __call__(self):
        if len(self.chunks) == 1: # fastpath
            return 

        # set value reference (accessing filtered_text does the work)
        self.event.value = self.filtered_text

    def _filterCore(self,  chunk, **kwargs):
        """Subclasses override this to provide specific impls"""
        return ''


def ijoin(a,b):
    """yield a0,b0,a1,b1.. if len(a) = len(b)+1"""
    yield(a[0])
    for i in range(1,len(a)):
        yield(b[i-1])
        yield(a[i])



import os, sys, time
import unittest
from sets import Set
import traceback

from Testing import ZopeTestCase

from wicked.normalize import titleToNormalizedId
from wicked.config import BACKLINK_RELATIONSHIP
from wicked.atcontent.ironicwiki import IronicWiki
from wickedtestcase import WickedTestCase, makeContent, WickedSite

from zope.component import adapter, provideAdapter
from zope.interface import implementer, Interface
from zope.interface import alsoProvides, noLongerProvides
from wicked.interfaces import IScope, IAmWicked, IAmWickedField

class IScopedField(IAmWickedField):
    """ scope marker """


@implementer(IScope)
@adapter(IScopedField, IAmWicked)
def scopeTester(field, context):
    """
    a demo scope function for testing, returns a scope that includes
    the parent of the current container, but not anything two levels
    (or more) up
    """
    scope_obj = context.aq_inner.aq_parent
    if not context.isPrincipiaFolderish:
        scope_obj = scope_obj.aq_inner.aq_parent
    path = '/'.join(scope_obj.getPhysicalPath())
    return path


class Scoped(WickedSite):
    
    @classmethod
    def setUp(cls):
        alsoProvides(IronicWiki.schema['body'], IScopedField)
        # @@ maybe make this local?
        provideAdapter(scopeTester)
        
    @classmethod
    def tearDown(cls):
        # clear out markers
        noLongerProvides(IronicWiki.schema['body'], IScopedField)        


class TestWickedScope(WickedTestCase):
    wicked_type = 'IronicWiki'
    wicked_field = 'body'
    layer = Scoped

    def test_registration(self):
        # test our test ;)
        from zope.component import getMultiAdapter
        assert getMultiAdapter((IronicWiki.schema['body'], self.page1), IScope) \
               == '/plone/Members'

    def test_insideScope(self):
        # @@ this test passes without "scope". bogus?
        f2 = makeContent(self.folder, 'f2', 'Folder')
        w1 = makeContent(f2, 'w1', 'IronicWiki',
                          title='W1 Title')
        
        self.set_text(w1, "((%s))" % self.page1.Title())
        self.failUnlessWickedLink(w1, self.page1)

    def test_outsideScope(self):
        f2 = makeContent(self.folder, 'f2', 'Folder')
        f3 = makeContent(f2, 'f3', 'Folder')
        w1 = makeContent(f3, 'w1', 'IronicWiki',
                          title='W1 Title')
        self.set_text(w1, "((%s))" % self.page1.Title())
        self.failIfWickedLink(w1, self.page1)
        

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestWickedScope))
    return suite



from Products.CMFCore.utils import getToolByName
from cStringIO import StringIO
from wicked.at import config
from wicked.at.config import REFERENCE_MANAGER

def configureWysiwyg(portal, out):
    props = getToolByName(portal, 'portal_properties')
    if not hasattr(props, 'site_properties'): # not plone
        return
    
    editors = props.site_properties.getProperty('available_editors')
    if "Kupu" in editors:
        # move it up in the list
        editors = list(editors)
        editors.remove('Kupu')
        editors = ['Kupu',] + editors
        props.site_properties._updateProperty('available_editors', editors)


def configureReferenceCatalog(portal, out):
    catalog = getToolByName(portal, REFERENCE_MANAGER)
    for indexName, indexType in (
        ('targetId', 'FieldIndex'),
        ('targetTitle', 'FieldIndex'),
        ('targetURL', 'FieldIndex'), ):

        try:
            catalog.addIndex(indexName, indexType, extra=None)
        except:
            pass
        try:
            catalog.addColumn(indexName)
        except:
            pass

        catalog.manage_reindexIndex(indexName)


def install(self):
    out = StringIO()
    configureReferenceCatalog(self, out)
    configureWysiwyg(self, out)
    reindex = None
    pc = getToolByName(self, 'portal_catalog')
    if not 'UID' in pc.schema():
        pc.addColumn('UID')
        reindex = True
        
    if not 'UID' in pc.indexes():
        pc.addIndex('UID', 'FieldIndex')
        reindex = True
        
    pc.manage_reindexIndex('UID')
    print >> out, "Successfully installed %s." % config.PROJECTNAME
    return out.getvalue()



from Products.Archetypes import public as atapi
from Products.CMFCore.permissions import setDefaultRoles
import Products.Archetypes.public as atapi

PROJECTNAME='wicked.atcontent'

import migration

def initialize(context):
    """load up content types"""
    from Products.CMFCore import utils as cmf_utils
    #app = context._ProductContext__app
    #patch_listDefaultTypeInformation(app)
    
    from ironicwiki import IronicWiki
    try:
        from wickeddoc import WickedDoc
    except ImportError: # no ATCT
        pass

    types = atapi.listTypes(PROJECTNAME)
    content_types, constructors, ftis = atapi.process_types( types,
                                                             PROJECTNAME)
    permissions = {}
    types = atapi.listTypes(PROJECTNAME)

    for atype in  types:
        permission = "%s: Add %s" % (PROJECTNAME, atype['portal_type'])
        permissions[atype['portal_type']] = permission

        # Assign default roles
        setDefaultRoles(permission, ('Manager', 'Owner'))
    
    allTypes = zip(content_types, constructors)
    for atype, constructor in allTypes:
        kind = "%s: %s" % (PROJECTNAME, atype.archetype_name)
        cmf_utils.ContentInit(
            kind,
            content_types      = (atype,),
            permission         = permissions[atype.portal_type],
            extra_constructors = (constructor,),
            fti                = ftis,
            ).initialize(context)



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit@kalistra.com>
#   - and contributors
#
##########################################################

"""
BackLink
~~~~~~~~

A type of reference that can be used/extended to provide smarter
inter-document linking. With the support of an editor this can be
quite useful.
"""
from Products.Archetypes.references import Reference
from wicked import config
from wicked.interfaces import IWickedBacklink
from zope.interface import implements

class Backlink(Reference):
    """
    A backlink is a reference set on an object when it is targetted
    by a resolved wicked-link
    """
    implements(IWickedBacklink)
    relationship = config.BACKLINK_RELATIONSHIP
    def __repr__(self):
        return "<Backlink sid:%s tid:%s rel:%s>" %(self.sourceUID, self.targetUID, self.relationship)

    def targetURL(self):
        """
        let's stick this in the catalog this to keep things light
        """
        target = self.getTargetObject()
        if target:
            return target.absolute_url()
        return '#'


        



from cStringIO import StringIO
from Products.Archetypes.public import listTypes
from Products.Archetypes.Extensions.utils import installTypes
from wicked.atcontent import zope2
from wicked.at.Extensions.Install import install as installwickedat

def install(portal):
    out = StringIO()
    installwickedat(portal)
    installTypes(portal, out, listTypes(zope2.PROJECTNAME), zope2.PROJECTNAME)




from setuptools import setup, find_packages
import sys, os

version = '1.1.1'

setup(name='wicked',
      version=version,
      description="wicked is a compact syntax for doing wiki-like content linking and creation",
      long_description="""\
""",
      classifiers=[
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Framework :: Plone",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='wiki anti-wiki zope3 zope2 plone',
      author='whit',
      author_email='wicked@lists.openplans.org',
      url='http://openplans.org/projects/wicked',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['wicked'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[],
      entry_points="""
      # -*- Entry points: -*-
      [wicked.base_registration]
      basic_plone_registration = wicked.registration:BasePloneWickedRegistration
      bracketted_plone_registration = wicked.registration:BasePloneMediaWickedRegistration
      #[wicked.cache_manager]
      """,
      )



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from urllib import quote
from zope.interface import Interface
from zope.component import adapter, handle
from Products.Five import BrowserView
from wicked import utils
from wicked.normalize import titleToNormalizedId as normalize
from wicked.config import BACKLINK_RELATIONSHIP
from wicked.utils import getWicked
from wicked.link import WickedAdd, BasicLink
from wicked.interfaces import IWickedContentAddedEvent
from Products.Archetypes.interfaces import IBaseObject

_marker=object()

class ATWickedAdd(WickedAdd, BrowserView):

    def add_content(self, title=None, section=None, type_name=_marker):
        # make it possible to pass in container
        # pre-populate with backlinks?
        type_name = self.request.get('type_name', type_name)
        if type_name is _marker:
            type_name = self.context.portal_type
        title = self.request.get('Title', title)
        section = self.request.get('section', section)
        assert title, 'Must have a title to create content' 
        newcontentid = normalize(title)

        container = self.aq_parent.aq_parent
        container.invokeFactory(type_name, id=newcontentid,
                                   title=title)

        newcontent = getattr(self.context, newcontentid)
        self.notify_content_added(newcontent, title, section)

# channel on AT
@adapter(IBaseObject, IWickedContentAddedEvent)
def handle_at_newcontent(context, event):
    field = context.Schema()[event.section]
    wicked = getWicked(field, context, event)
    wicked.manageLink(context, event.title)
    portal_status_message='"%s" has been created' % event.title
    url = event.newcontent.absolute_url()
    restr = "%s/edit?portal_status_message=%s" %(url, quote(portal_status_message))
    event.request.RESPONSE.redirect(restr)

    
class BasicFiveLink(BasicLink, BrowserView):
    """
    Five prepared link implementation
    """
    __init__=BasicLink.__init__


def test_suite():
    import unittest
    from zope.testing import doctest
    optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS
    from Products.PloneTestCase.layer import PloneSite, ZCMLLayer
    from Testing.ZopeTestCase import ZopeDocFileSuite, FunctionalDocFileSuite
    from Products.PloneTestCase import ptc
    ptc.setupPloneSite()
    renderer = ZopeDocFileSuite('renderer.txt',
                                package='wicked.at',
                                optionflags=optionflags)
    setattr(renderer, 'layer', ZCMLLayer)
    add = FunctionalDocFileSuite('add.txt',
                           package='wicked.at',
                           test_class=ptc.FunctionalTestCase,                                   
                           optionflags=optionflags)
                 
    return unittest.TestSuite((add, renderer))





from zope.interface import Interface, Attribute, implements
from wicked.fieldevent.interfaces import ITxtFilter, IFieldEvent
from wicked.fieldevent.interfaces import IFieldStorageEvent
from zope.annotation.interfaces import IAnnotatable

class IUID(Interface):
    """an effectively unique identifier for the context of the system"""


class IWickedFilter(ITxtFilter):
    """
    Wicked txtfilter

    @@ fill out interface for subscription adapter
    """


class IAmWickedField(Interface):
    """
    A 'field' specifically designated for wicked behavior.  Could be
    just a wrapper for a getter and a setter combo.

    @@ will probably initially have some at cruft to it
    """


class IAmWicked(IAnnotatable):
    """
    wicked content... marker to indicate a piece of content has wicked
    fields
    """


class IWickedTarget(Interface):
    """
    marker interface for an object linked to in a wicked text area
    """


class IWickedBacklink(Interface):
    """
    Backlink marker

    @@ is this used?
    """

    
class IWickedFilter(Interface):
    """
    Wicked resolving filter
    """
    ### this need complete documentation and test to verify


class IBacklinkManager(Interface):
    """
    this might become the wicked storage manager...
    """
    def manageLinks():
        """
        removes old links adds new
        """

    def addLinks(links, scope, dups=[]):
        """
        retrieves brains, and sets backlinks for a colletion
        wicked links.  will filter against identifiers(UIDs)
        in dups (this allow chaining with removelinks, which
        returns duplicate links)
        """
        
    def getLinks():
        """
        returns all current backlinks
        """
        
    def set(brain, link):
        """
        creates a backlink(a smart pointer pointing for the links target)
        and caches link
        """
        
    def remove(brain):
        """
        does the actual backlink removal and cache unsetting
        """

    def removeLinks(exclude=tuple()):
        """
        iterates over a list of brain representing existing backlink
        and executes backlink deletion if not included in exclude
        
        @exclude: list of strings 'links' not to erase
        """

class IATBacklinkManager(IBacklinkManager):
    """
    A manager for Archetypes reference (aka smart pointer) based backlinka 
    """
    relation = Attribute( """
    Name of Archetype relationship. Used to retrieve
    backlinks from reference engine
    """)


class ICacheManager(Interface):
    """
    a cache manager that cache per content instance
    on the content type itself. It manages a dual level data
    structure: dict(section=dict(key=value))
    """
    cache_attr = Attribute('attribute to access cache through')
    section = Attribute('attribute to access sub-cache through')
    cache=Attribute('Actually persisted cache object')

    def __init__(field, context):
        """
        cache manager multiadapts the field and the context
        """

    def get(key, default, **kwargs):
        """
        @rendered wicked link
        """
        
    def set(key, text):
        """
        sets text to cache
        """
        
    def unset(key):
        """
        invalidates cache key
        """

    def _get_store():
        """
        create and / or
        @return the  master store
        """
        
    def _get_cache():
        """
        @return actual datastructure
        for getting and setting
        """
        
class IWickedQuery(Interface):
    """
    object for handling and returning
    dataobjects for wicked prep
    for the macro parser
    """
    
    chunk = Attribute('unaltered string from inside ((wicked link))')
    normalized = Attribute('normalled chunk')
    scope = Attribute('scoping parameter for "scoped" searches')

    def scopedSearch(best_match):
        """
        @param best_match : attempt to make
        best match for query returned

        @return: list of dataobjects
        """
        
    def search(chunk, normalized, best_match):
        """
        @param best_match : attempt to make
        best match for query returned

        @return : list of dataobjects
        """
        
    def configure(chunk, normalized, scope):
        """
        configure set instance attributes
        
        @param chunk : instance attr 
        @param normalized : instance attr
        @param scope : instance attr

        @return : None
        """
        
class IWickedLink(Interface):
    """renderer for wicked links"""

    def context():
        """
        context
        """
        
    def howmany():
        """
        @return integer
        how many links
        """
        
    def multiple():
        """
        @return boolean
        howmany > 1
        """

    def links():
        """
        @return list
        list of link datum
        """

    def singlelink():
        """
        @return boolean
        howmany == 1
        """

    def load(links, chunk, section):
        """
        load data for repeated rendering
        """


class IWickedEvent(Interface):
    """something wicked is happening"""


class IValueToString(Interface):
    """converts a field object to a raw string. mainly used to get
    around AT's un normalized value handling"""


class IScope(Interface):
    """a value used to determine breadth of resolver's search. usually a path"""


class WickedEvent(object):
    implements(IWickedEvent)

class IWickedContentAddedEvent(Interface):
    """a piece of content is add by wicked"""
    title = Attribute("title of new content")
    section = Attribute("subobject of concern")
    request = Attribute("current request")
    section = Attribute("field name inside context")
    context = Attribute("context linked from")
    newcontent = Attribute("object representing new content")

class WickedContentAddedEvent(WickedEvent):
    implements(IWickedContentAddedEvent)
    
    def __init__(self, context, newcontent, title, section, request=None):
        self.__dict__.update(locals())



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit@kalistra.com>
#   - Rob Miller <rob@kalistra.com> (RaFromBRC)
#   - and contributors
#
##########################################################
from interfaces import IAmWickedField, IAmWicked, IFieldEvent 
from interfaces import ICacheManager, IValueToString, IScope
from interfaces import IWickedFilter, IWickedQuery, IBacklinkManager
from normalize import titleToNormalizedId as normalize
from wicked import utils
from wicked.fieldevent.interfaces import EndFiltrationException
from wicked.fieldevent.interfaces import ITxtFilterList, IFieldRenderEvent
from wicked.fieldevent.interfaces import IFieldValueSetter, IFieldStorageEvent
from zope.component.interfaces import ComponentLookupError
from wicked.fieldevent.txtfilter import TxtFilter
from zope.component import getMultiAdapter, adapts, adapter
from zope.interface import implements, implementer, Interface

import sre

_marker = object()

pattern = sre.compile(r'\(\(([\w\W]+?)\)\)') # matches ((Some Text To link 123))

class WickedFilter(TxtFilter):
    implements(IWickedFilter)
    adapts(IAmWickedField, IAmWicked, IFieldEvent)

    name = 'Wicked Filter'

    pattern = pattern
    query_iface = IWickedQuery

    def __init__(self, field, instance, event):
        super(WickedFilter, self).__init__(field, instance, event)
        self.section = field.__name__

    @utils.memoizedproperty
    def scope(self):
        try:
            return getMultiAdapter((self.field, self.context), IScope)
        except ComponentLookupError:
            return ''

    # avoid global lookup
    getMatch = staticmethod(utils.getMatch)
    _normalize = staticmethod(normalize)

    # optimization
    @utils.memoize
    def normalize(self, value):
        return self._normalize(value)
    
    def _filterCore(self, chunk, **kwargs):
        normalled = self.normalize(chunk)
        links=self.getLinks(chunk, normalled)
        self.renderer.load(links, chunk)
        return self.renderer()

    @property
    def filtered_text(self):
        """syntax preprocessing"""
        return super(WickedFilter, self).filtered_text
    
    @utils. memoize
    @utils.linkcache
    def getLinks(self, chunk, normalled):
        self.resolver.configure(chunk, normalled, self.scope)
        brains = self.resolver.search
        if not brains:
            brains = self.resolver.scopedSearch
        links = [utils.packBrain(b) for b in brains if b]
        return links

    @utils.memoizedproperty
    def resolver(self):
        """
        @return query object
        """
        return self.query_iface(self.context)

    @utils.memoizedproperty    
    def backlinker(self):
        return getMultiAdapter((self, self.context), IBacklinkManager)

    def manageLink(self, obj, link):
        self.backlinker.manageLink(obj, link)

    def unlink(self, uid):
        self.backlinker.unlink(uid)
        
    def manageLinks(self, links):
        self.backlinker.manageLinks(links)
        
    @utils.memoizedproperty
    def cache(self):
        return getMultiAdapter((self, self.context), ICacheManager)
    
    @utils.memoizedproperty
    def renderer(self):
        # @@ better way to get request?
        renderer = getMultiAdapter((self.context, self.context.REQUEST), Interface, 'link_renderer')
        renderer.section = self.section
        # hook for zope2 aq wrapper
        if hasattr(renderer, '__of__'):
            return renderer.__of__(self.context)
        return render

    def __call__(self):
        if self.event.kwargs.get('raw', False):
            raise EndFiltrationException('Kwargs flag for raw return')
        super(WickedFilter, self).__call__()

    def removeParens(wikilink):
        wikilink.replace('((', '')
        wikilink.replace('))', '')
        return wikilink
    removeParens=staticmethod(removeParens)


class BrackettedWickedFilter(WickedFilter):
    """media wiki style bracket matching"""
    pattern=sre.compile(r'\[\[([\w\W]+?)\]\]') # matches [[Some Text To link 123]]
    def removeParens(wikilink):
        wikilink.replace('[[', '')
        wikilink.replace(']]', '')
        return wikilink
    removeParens=staticmethod(removeParens)

NAME = WickedFilter.name


## event handlers ##

def wicked_listener(field, instance, event):
    """standalone wicked filter (ie not as a txtfilter). Optimal if
    not using txtfilters"""
    
    if event.kwargs.get('raw', False):
        return # no processing (probably should be caught even earlier)
    
    wicked = getMultiAdapter((field, instance, event), IWickedFilter)
    try:
        wicked()
    except EndFiltrationException:
        pass


@implementer(IFieldValueSetter)
@adapter(IAmWickedField, IFieldStorageEvent)
def backlink_handler(field, event):
    try:
        wicked = utils.getWicked(field, event.instance, event)
    except ComponentLookupError:
        # no adapter registered for this type currently
        # @@ This might be handle better by redispatch
        return
        
    if not event.value:
        return

    value = event.value
    value_str = value

    try:
        # this block handle conversions for file uploads and
        # atapi.BaseUnit or any other not quite plain text "value objects"
        value_str = getMultiAdapter((value, field), IValueToString)
    except ComponentLookupError:
        pass


    found = wicked.pattern.findall(value_str)

    if not len(found):
        return

    new_links = [wicked.removeParens(link) for link in found]
    wicked.manageLinks(new_links)


class BacklinkRegistrationProxy(object):
    """cuz you can't pickle functions with decorators"""
    implements(IFieldValueSetter)
    adapts(IAmWickedField, IFieldStorageEvent)
    
    __init__ = staticmethod(backlink_handler)

## toy example code ##
    
@adapter(IAmWickedField, IAmWicked, IFieldRenderEvent)
@implementer(ITxtFilterList)
def filter_list(field, context, event):
    """example adapter for a one item list for ordering a txtfilter
    pipeline involving wicked only.  Practically useless, for example
    only"""
    return [NAME]



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit at www.openplans.org>
#   - and contributors
#
##########################################################
from wicked import config, utils
from zope.app.container.interfaces import IObjectRemovedEvent, IObjectAddedEvent
from wicked.interfaces import IAmWicked, IAmWickedField, IUID
from Products.Archetypes.interfaces import ISchema

# @@ these a very at specific

def at_handle_target_deletion(ref, event):
    """
    Invalidate any pointer before object deletion
    """
    target = ref.getTargetObject()
    for field in ISchema(target).fields():
        if IAmWickedField.providedBy(field):
            wicked = utils.getWicked(field, target)
            wicked.unlink(ref.sourceUID)
            break

def at_handle_target_moved(obj, event):
    """
    when a target of a link is moved, or renamed we need to notify any
    objects that may be caching pointers
    """
    #@@ add more tests
    if IObjectRemovedEvent.providedBy(event):
        return
    
    refs=obj.getRefs(relationship=config.BACKLINK_RELATIONSHIP)
    path = '/'.join(obj.getPhysicalPath())
    for target in refs:
        for field in ISchema(target).fields():
            wicked = utils.getWicked(field, ref)
            uid = IUID(obj)
            data = dict(path=path,
                        icon=obj.getIcon(),
                        uid=uid)
            wicked.cache.reset(uid, [data])

    
